#pragma once
#include <stdint.h>

extern const uint8_t bssmap_reset[18];
extern const uint8_t bssmap_reset_ack[19];
extern const uint8_t bssmap_paging[32];
extern const uint8_t bssmap_udt[28];
extern const uint8_t bssmap_cr[44];
extern const uint8_t bssmap_cc[10];
extern const uint8_t bssmap_dtap[22];
extern const uint8_t bssmap_clear[13];
extern const uint8_t bssmap_released[14];
extern const uint8_t bssmap_release_complete[7];
extern const uint8_t tcap_global_title[183];
